@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: Stg.h,v 1.38 2001/08/29 17:24:25 qrczak Exp $
+ * $Id: Stg.h,v 1.55.2.1 2003/09/22 09:14:12 simonmar Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -7,6 +7,11 @@
  *
  * This file is included *automatically* by all .hc files.
  *
+ * NOTE: always include Stg.h *before* any other headers, because we
+ * define some register variables which must be done before any inline
+ * functions are defined (some system headers have been known to
+ * define the odd inline function).
+ *
  * ---------------------------------------------------------------------------*/
 
 #ifndef STG_H
@@ -31,6 +36,10 @@
 /* Configuration */
 #include "config.h"
 
+/* This needs to be up near the top as the register line on alpha needs
+ * to be before all procedures */
+#include "TailCalls.h"
+
 #if __GNUC__ >= 3
 /* Assume that a flexible array member at the end of a struct
  * can be defined thus: T arr[]; */
@@ -40,6 +49,10 @@
 #define FLEXIBLE_ARRAY 0
 #endif
 
+#if defined(SMP) || defined(THREADED_RTS)
+#define RTS_SUPPORTS_THREADS 1
+#endif
+
 /* Some macros to handle DLLing (Win32 only at the moment). */
 #include "StgDLL.h"
 
@@ -69,15 +82,12 @@
 #  define LAZY_BLACKHOLING
 #endif
 
-/* ToDo: remove */
-#define COMPILER 1
-
 /* TABLES_NEXT_TO_CODE says whether to assume that info tables are
  * assumed to reside just before the code for a function.
  *
  * UNDEFINING THIS WON'T WORK ON ITS OWN.  You have been warned.
  */
-#ifndef USE_MINIINTERPRETER
+#if !defined(USE_MINIINTERPRETER) && !defined(ia64_TARGET_ARCH)
 #define TABLES_NEXT_TO_CODE
 #endif
 
@@ -103,6 +113,14 @@ void _stgAssert (char *, unsigned int);
 	    _stgAssert(__FILE__, __LINE__)
 #endif /* DEBUG */
 
+/* 
+ * Use this on the RHS of macros which expand to nothing
+ * to make sure that the macro can be used in a context which
+ * demands a non-empty statement.
+ */
+
+#define doNothing() do { } while (0)
+
 /* -----------------------------------------------------------------------------
    Global type definitions
    -------------------------------------------------------------------------- */
@@ -150,8 +168,10 @@ typedef StgWord64       LW_;
 
 /* Profiling information */
 #include "StgProf.h"
+#include "StgLdvProf.h"
 
 /* Storage format definitions */
+#include "StgFun.h"
 #include "Closures.h"
 #include "ClosureTypes.h"
 #include "InfoTables.h"
@@ -167,27 +187,17 @@ typedef StgWord64       LW_;
 #include "SMP.h"
 #include "MachRegs.h"
 #include "Regs.h"
-#include "TailCalls.h"
 #include "Block.h"
 
 /* RTS public interface */
 #include "RtsAPI.h"
 
-/* these are all ANSI C headers */
+/* System headers: stdlib.h is eeded so that we can use NULL.  It must
+ * come after MachRegs.h, because stdlib.h might define some inline
+ * functions which may only be defined after register variables have
+ * been declared.
+ */
 #include <stdlib.h>
-#include <string.h>
-#include <math.h>
-#include <assert.h>
-#include <errno.h>
-#include <stdio.h>
-
-#ifdef HAVE_SIGNAL_H
-#include <signal.h>
-#endif
-
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
 
 #ifdef SMP
 #include <pthread.h>
@@ -215,16 +225,21 @@ typedef StgWord64       LW_;
 /* Runtime-system hooks */
 #include "Hooks.h"
 
+#include "Signals.h"
+
 #include "HsFFI.h"
 
 /* Misc stuff without a home */
 DLL_IMPORT_RTS extern char **prog_argv;	/* so we can get at these from Haskell */
 DLL_IMPORT_RTS extern int    prog_argc;
-
-extern char **environ;
+DLL_IMPORT_RTS extern char  *prog_name;
 
 extern void stackOverflow(void);
 
+#if defined(WANT_DOTNET_SUPPORT)
+#include "DNInvoke.h"
+#endif
+
 /* Creating and destroying an adjustor thunk.
    I cannot make myself create a separate .h file
    for these two (sof.) 
