@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: BlockAlloc.c,v 1.10.4.1 2001/12/13 13:37:15 simonmar Exp $
+ * $Id: BlockAlloc.c,v 1.16 2003/02/18 05:47:53 sof Exp $
  *
  * (c) The GHC Team 1998-2000
  * 
@@ -23,11 +23,13 @@
 #include "BlockAlloc.h"
 #include "MBlock.h"
 
+#include <string.h>
+
 static void    initMBlock(void *mblock);
 static bdescr *allocMegaGroup(nat mblocks);
 static void    freeMegaGroup(bdescr *bd);
 
-static bdescr *free_list;
+static bdescr *free_list = NULL;
 
 /* -----------------------------------------------------------------------------
    Initialisation
@@ -35,7 +37,7 @@ static bdescr *free_list;
 
 void initBlockAllocator(void)
 {
-  free_list = NULL;
+    // The free list starts off NULL
 }
 
 /* -----------------------------------------------------------------------------
@@ -263,7 +265,7 @@ freeMegaGroup(bdescr *p)
 
   n = p->blocks * BLOCK_SIZE / MBLOCK_SIZE + 1;
   for (; n > 0; (W_)p += MBLOCK_SIZE, n--) {
-    initMBlock((void *)((W_)p & ~MBLOCK_MASK));
+    initMBlock(MBLOCK_ROUND_DOWN(p));
     initGroup(BLOCKS_PER_MBLOCK, p);
     freeGroup(p);
   }
