@@ -1,7 +1,7 @@
 /* -----------------------------------------------------------------------------
- * $Id: RtsStartup.c,v 1.52.4.2 2001/12/13 14:21:21 simonmar Exp $
+ * $Id: RtsStartup.c,v 1.75.2.2 2004/02/27 13:47:35 simonmar Exp $
  *
- * (c) The GHC Team, 1998-2000
+ * (c) The GHC Team, 1998-2002
  *
  * Main function for a standalone Haskell program.
  *
@@ -17,7 +17,7 @@
 #include "Schedule.h"   /* initScheduler */
 #include "Stats.h"      /* initStats */
 #include "Signals.h"
-#include "Itimer.h"
+#include "Timer.h"      /* startTimer, stopTimer */
 #include "Weak.h"
 #include "Ticky.h"
 #include "StgRun.h"
@@ -25,6 +25,7 @@
 #include "Prelude.h"		/* fixupRTStoPreludeRefs */
 #include "HsFFI.h"
 #include "Linker.h"
+#include "ThreadLabels.h"
 
 #if defined(RTS_GTK_FRONTPANEL)
 #include "FrontPanel.h"
@@ -33,6 +34,7 @@
 #if defined(PROFILING) || defined(DEBUG)
 # include "Profiling.h"
 # include "ProfHeap.h"
+# include "RetainerProfile.h"
 #endif
 
 #if defined(GRAN)
@@ -48,52 +50,48 @@
 # include "LLC.h"
 #endif
 
-/*
- * Flag Structure
- */
-struct RTS_FLAGS RtsFlags;
-
-static int rts_has_started_up = 0;
-#if defined(PAR)
-ullong startTime = 0;
+#if defined(mingw32_TARGET_OS)
+#include "win32/AsyncIO.h"
 #endif
 
-EXTFUN(__stginit_Prelude);
-static void initModules ( void (*)(void) );
+#include <stdlib.h>
+
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+
+#ifdef HAVE_TERMIOS_H
+#include <termios.h>
+#endif
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#endif
+
+// Flag Structure
+struct RTS_FLAGS RtsFlags;
+
+// Count of how many outstanding hs_init()s there have been.
+static int hs_init_count = 0;
+
+#if HAVE_TERMIOS_H
+// Here we save the terminal settings on the standard file
+// descriptors, if we need to change them (eg. to support NoBuffering
+// input).
+struct termios *saved_termios[3] = {NULL,NULL,NULL};
+#endif
+
+/* -----------------------------------------------------------------------------
+   Starting up the RTS
+   -------------------------------------------------------------------------- */
 
 void
-setProgArgv(int argc, char *argv[])
+hs_init(int *argc, char **argv[])
 {
-   /* Usually this is done by startupHaskell, so we don't need to call this. 
-      However, sometimes Hugs wants to change the arguments which Haskell
-      getArgs >>= ... will be fed.  So you can do that by calling here
-      _after_ calling startupHaskell.
-   */
-   prog_argc = argc;
-   prog_argv = argv;
-}
-
-void
-getProgArgv(int *argc, char **argv[])
-{
-   *argc = prog_argc;
-   *argv = prog_argv;
-}
-
-
-void
-startupHaskell(int argc, char *argv[], void (*init_root)(void))
-{
-   /* To avoid repeated initialisations of the RTS */
-  if (rts_has_started_up) {
-    /* RTS is up and running, so only run the per-module initialisation code */
-    if (init_root) {
-      initModules(init_root);
+    hs_init_count++;
+    if (hs_init_count > 1) {
+	// second and subsequent inits are ignored
+	return;
     }
-    return;
-  } else {
-    rts_has_started_up=1;
-  }
 
     /* The very first thing we do is grab the start time...just in case we're
      * collecting timing statistics.
@@ -123,9 +121,11 @@ startupHaskell(int argc, char *argv[], void (*init_root)(void))
     defaultsHook();
 
     /* Parse the flags, separating the RTS flags from the programs args */
-    setupRtsFlags(&argc, argv, &rts_argc, rts_argv);
-    prog_argc = argc;
-    prog_argv = argv;
+    if (argc != NULL && argv != NULL) {
+	setupRtsFlags(argc, *argv, &rts_argc, rts_argv);
+	prog_argc = *argc;
+	prog_argv = *argv;
+    }
 
 #if defined(PAR)
     /* NB: this really must be done after processing the RTS flags */
@@ -155,46 +155,78 @@ startupHaskell(int argc, char *argv[], void (*init_root)(void))
     /* initialise the stable pointer table */
     initStablePtrTable();
 
+    /* initialise thread label table (tso->char*) */
+    initThreadLabelTable();
+
 #if defined(PROFILING) || defined(DEBUG)
     initProfiling1();
 #endif
 
-    /* run the per-module initialisation code */
-    initModules(init_root);
-
-#if defined(PROFILING) || defined(DEBUG)
-    initProfiling2();
-#endif
-
-    /* start the ticker */
-    install_vtalrm_handler();
-    initialize_virtual_timer(TICK_MILLISECS);
-
-    /* start our haskell execution tasks */
-#ifdef SMP
-    startTasks();
-#endif
+    /* start the virtual timer 'subsystem'. */
+    startTimer(TICK_MILLISECS);
 
     /* Initialise the stats department */
     initStats();
 
-#if !defined(mingw32_TARGET_OS) && !defined(PAR)
+#if defined(RTS_USER_SIGNALS)
     /* Initialise the user signal handler set */
     initUserSignals();
     /* Set up handler to run on SIGINT, etc. */
     initDefaultHandlers();
 #endif
  
+#if defined(mingw32_TARGET_OS)
+    startupAsyncIO();
+#endif
+
 #ifdef RTS_GTK_FRONTPANEL
     if (RtsFlags.GcFlags.frontpanel) {
 	initFrontPanel();
     }
 #endif
 
+#ifdef HAVE_LOCALE_H
+    setlocale(LC_ALL,"");
+#endif
+
     /* Record initialization times */
     stat_endInit();
 }
 
+// Compatibility interface
+void
+startupHaskell(int argc, char *argv[], void (*init_root)(void))
+{
+    hs_init(&argc, &argv);
+    hs_add_root(init_root);
+}
+
+
+/* -----------------------------------------------------------------------------
+   Getting/Setting the program's arguments.
+
+   These are used by System.Environment.
+   -------------------------------------------------------------------------- */
+
+void
+getProgArgv(int *argc, char **argv[])
+{
+    if (argc) { *argc = prog_argc; }
+    if (argv) { *argv = prog_argv; }
+}
+
+void
+setProgArgv(int argc, char *argv[])
+{
+   /* Usually this is done by startupHaskell, so we don't need to call this. 
+      However, sometimes Hugs wants to change the arguments which Haskell
+      getArgs >>= ... will be fed.  So you can do that by calling here
+      _after_ calling startupHaskell.
+   */
+   prog_argc = argc;
+   prog_argv = argv;
+}
+
 /* -----------------------------------------------------------------------------
    Per-module initialisation
 
@@ -213,8 +245,7 @@ startupHaskell(int argc, char *argv[], void (*init_root)(void))
 
    The code generator inserts a small function "__stginit_<module>" in each
    module and calls the registration functions in each of the modules it
-   imports.  So, if we call "__stginit_PrelMain", each reachable module in the
-   program will be registered (because PrelMain.mainIO calls Main.main).
+   imports.
 
    The init* functions are compiled in the same way as STG code,
    i.e. without normal C call/return conventions.  Hence we must use
@@ -223,117 +254,171 @@ startupHaskell(int argc, char *argv[], void (*init_root)(void))
 
 /* The init functions use an explicit stack... 
  */
-#define INIT_STACK_SIZE  (BLOCK_SIZE * 4)
-F_ *init_stack = NULL;
-nat init_sp = 0;
+#define INIT_STACK_BLOCKS  4
+static F_ *init_stack = NULL;
 
-static void
-initModules ( void (*init_root)(void) )
+void
+hs_add_root(void (*init_root)(void))
 {
+    bdescr *bd;
 #ifdef SMP
     Capability cap;
 #else
 #define cap MainCapability
 #endif
+    nat init_sp;
+
+    if (hs_init_count <= 0) {
+	barf("hs_add_root() must be called after hs_init()");
+    }
 
     init_sp = 0;
-    init_stack = (F_ *)allocate(INIT_STACK_SIZE / sizeof(W_));
+    bd = allocGroup(INIT_STACK_BLOCKS);
+    init_stack = (F_ *)bd->start;
     init_stack[init_sp++] = (F_)stg_init_ret;
-    init_stack[init_sp++] = (F_)__stginit_Prelude;
     if (init_root != NULL) {
 	init_stack[init_sp++] = (F_)init_root;
     }
     
     cap.r.rSp = (P_)(init_stack + init_sp);
     StgRun((StgFunPtr)stg_init, &cap.r);
+
+    freeGroup(bd);
+
+#if defined(PROFILING) || defined(DEBUG)
+    // This must be done after module initialisation.
+    // ToDo: make this work in the presence of multiple hs_add_root()s.
+    initProfiling2();
+#endif
 }
 
 /* -----------------------------------------------------------------------------
- * Shutting down the RTS - two ways of doing this, one which
- * calls exit(), one that doesn't.
- *
- * (shutdownHaskellAndExit() is called by System.exitWith).
- * -----------------------------------------------------------------------------
- */
-void
-shutdownHaskellAndExit(int n)
-{
-  OnExitHook();
-  shutdownHaskell();
-#if defined(PAR)
-  /* really exit (stg_exit() would call shutdownParallelSystem() again) */
-  exit(n);
-#else
-  stg_exit(n);
-#endif
-}
+   Shutting down the RTS
+   -------------------------------------------------------------------------- */
 
 void
-shutdownHaskell(void)
+hs_exit(void)
 {
-  if (!rts_has_started_up)
-     return;
-
-  /* start timing the shutdown */
-  stat_startExit();
+    if (hs_init_count <= 0) {
+	barf("too many hs_exit()s");
+    }
+    hs_init_count--;
+    if (hs_init_count > 0) {
+	// ignore until it's the last one
+	return;
+    }
 
+    /* start timing the shutdown */
+    stat_startExit();
+    
 #if !defined(GRAN)
-  /* Finalize any remaining weak pointers */
-  finalizeWeakPointersNow();
+    /* Finalize any remaining weak pointers */
+    finalizeWeakPointersNow();
 #endif
-
+    
+    /* stop all running tasks */
+    exitScheduler();
+    
 #if defined(GRAN)
-  /* end_gr_simulation prints global stats if requested -- HWL */
-  if (!RtsFlags.GranFlags.GranSimStats.Suppressed)
-    end_gr_simulation();
+    /* end_gr_simulation prints global stats if requested -- HWL */
+    if (!RtsFlags.GranFlags.GranSimStats.Suppressed)
+	end_gr_simulation();
 #endif
 
-  /* stop all running tasks */
-  exitScheduler();
+    /* stop the ticker */
+    stopTimer();
+    
+    /* reset the standard file descriptors to blocking mode */
+    resetNonBlockingFd(0);
+    resetNonBlockingFd(1);
+    resetNonBlockingFd(2);
 
-  /* stop the ticker */
-  initialize_virtual_timer(0);
-  
-  /* reset the standard file descriptors to blocking mode */
-  resetNonBlockingFd(0);
-  resetNonBlockingFd(1);
-  resetNonBlockingFd(2);
+#if HAVE_TERMIOS_H
+    // Reset the terminal settings on the standard file descriptors,
+    // if we changed them.  See System.Posix.Internals.tcSetAttr for
+    // more details, including the reason we termporarily disable
+    // SIGTTOU here.
+    { 
+	int fd;
+	sigset_t sigset, old_sigset;
+	sigemptyset(&sigset);
+	sigaddset(&sigset, SIGTTOU);
+	sigprocmask(SIG_BLOCK, &sigset, &old_sigset);
+	for (fd = 0; fd <= 2; fd++) {
+	    if (saved_termios[fd] != NULL) {
+		tcsetattr(fd,TCSANOW,saved_termios[fd]);
+	    }
+	}
+	sigprocmask(SIG_SETMASK, &old_sigset, NULL);
+    }
+#endif
 
 #if defined(PAR)
-  /* controlled exit; good thread! */
-  shutdownParallelSystem(0);
-
-  /* global statistics in parallel system */
-  PAR_TICKY_PAR_END();
+    /* controlled exit; good thread! */
+    shutdownParallelSystem(0);
+    
+    /* global statistics in parallel system */
+    PAR_TICKY_PAR_END();
 #endif
 
-  /* stop timing the shutdown, we're about to print stats */
-  stat_endExit();
-
-  /* clean up things from the storage manager's point of view.
-   * also outputs the stats (+RTS -s) info.
-   */
-  exitStorage();
-
+    /* stop timing the shutdown, we're about to print stats */
+    stat_endExit();
+    
+    /* clean up things from the storage manager's point of view.
+     * also outputs the stats (+RTS -s) info.
+     */
+    exitStorage();
+    
 #ifdef RTS_GTK_FRONTPANEL
     if (RtsFlags.GcFlags.frontpanel) {
 	stopFrontPanel();
     }
 #endif
 
-#if defined(PROFILING) || defined(DEBUG)
-  endProfiling();
-#endif
-
 #if defined(PROFILING) 
-  report_ccs_profiling();
+    reportCCSProfiling();
 #endif
 
+#if defined(PROFILING) || defined(DEBUG)
+    endProfiling();
+#endif
+
+#ifdef PROFILING
+    // Originally, this was in report_ccs_profiling().  Now, retainer
+    // profiling might tack some extra stuff on to the end of this file
+    // during endProfiling().
+    fclose(prof_file);
+#endif
+    
 #if defined(TICKY_TICKY)
-  if (RtsFlags.TickyFlags.showTickyStats) PrintTickyInfo();
+    if (RtsFlags.TickyFlags.showTickyStats) PrintTickyInfo();
 #endif
 
-  rts_has_started_up=0;
+#if defined(mingw32_TARGET_OS)
+    shutdownAsyncIO();
+#endif
+}
+
+// Compatibility interfaces
+void
+shutdownHaskell(void)
+{
+    hs_exit();
+}
+
+void
+shutdownHaskellAndExit(int n)
+{
+    if (hs_init_count == 1) {
+	OnExitHook();
+	hs_exit();
+#if defined(PAR)
+	/* really exit (stg_exit() would call shutdownParallelSystem() again) */
+	exit(n);
+#else
+	stg_exit(n);
+#endif
+    }
 }
 
 /* 
@@ -345,7 +430,7 @@ static int exit_started=rtsFalse;
 #endif
 
 void  
-stg_exit(I_ n)
+stg_exit(int n)
 { 
 #ifdef PAR
   /* HACK: avoid a loop when exiting due to a stupid error */
