@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: RtsUtils.h,v 1.11.2.1 2001/12/13 14:22:31 simonmar Exp $
+ * $Id: RtsUtils.h,v 1.17.2.1 2004/02/26 16:36:50 simonmar Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -7,11 +7,14 @@
  *
  * ---------------------------------------------------------------------------*/
 
+#ifndef RTSUTILS_H
+#define RTSUTILS_H
+
+/* (Checked) dynamic allocation: */
 extern void *stgMallocBytes(int n, char *msg);
-extern void *stgMallocWords(int n, char *msg);
 extern void *stgReallocBytes(void *p, int n, char *msg);
-extern void *stgReallocWords(void *p, int n, char *msg);
 extern void *stgCallocBytes(int n, int m, char *msg);
+extern void stgFree(void* p);
 
 extern void barf(char *s, ...) __attribute__((__noreturn__)) ;
 extern void belch(char *s, ...);
@@ -27,11 +30,17 @@ extern void resetNonBlockingFd(int fd);
 extern nat stg_strlen(char *str);
 
 /*Defined in Main.c, but made visible here*/
-extern void stg_exit(I_ n) __attribute__((noreturn));
+extern void stg_exit(int n) __attribute__((noreturn));
 
 char *time_str(void);
 char *ullong_format_string(ullong, char *, rtsBool);
 
+#ifdef PAR
 ullong   msTime(void);
+#endif
 
+#ifdef DEBUG
+extern void heapCheckFail( void );
+#endif
 
+#endif // RTSUTILS_H
