@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: Stats.h,v 1.11 2001/07/23 17:23:20 simonmar Exp $
+ * $Id: Stats.h,v 1.17 2002/02/06 01:21:41 sof Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -14,6 +14,20 @@ extern void      stat_startGC(void);
 extern void      stat_endGC(lnat alloc, lnat collect, lnat live, 
 			    lnat copied, lnat gen);
 
+#ifdef PROFILING
+extern void      stat_startRP(void);
+extern void      stat_endRP(nat, 
+#ifdef DEBUG_RETAINER
+                            nat, int, 
+#endif
+                            double);
+#endif // PROFILING
+
+#if defined(PROFILING) || defined(DEBUG)
+extern void      stat_startHeapCensus(void);
+extern void      stat_endHeapCensus(void);
+#endif
+
 extern void      stat_startExit(void);
 extern void      stat_endExit(void);
 
@@ -25,5 +39,14 @@ extern void      initStats(void);
 extern double    mut_user_time_during_GC(void);
 extern double    mut_user_time(void);
 
+#ifdef PROFILING
+extern double    mut_user_time_during_RP(void);
+extern double    mut_user_time_during_heap_census(void);
+#endif // PROFILING
+
 extern void      statDescribeGens( void );
 extern HsInt     getAllocations( void );
+
+#if defined(SMP)
+extern long int  stat_getElapsedTime ( void );
+#endif
