@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: StoragePriv.h,v 1.17.4.2 2001/12/13 14:43:13 simonmar Exp $
+ * $Id: StoragePriv.h,v 1.24 2002/12/19 14:33:23 simonmar Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -21,6 +21,7 @@ extern step *g0s0;
 extern generation *oldest_gen;
 
 extern void newCAF(StgClosure*);
+extern void newDynCAF(StgClosure *);
 
 extern void move_TSO(StgTSO *src, StgTSO *dest);
 extern StgTSO *relocate_stack(StgTSO *dest, ptrdiff_t diff);
@@ -28,9 +29,13 @@ extern StgTSO *relocate_stack(StgTSO *dest, ptrdiff_t diff);
 extern StgClosure *static_objects;
 extern StgClosure *scavenged_static_objects;
 
+extern StgWeak *old_weak_ptr_list;
+
 extern StgWeak    *weak_ptr_list;
 extern StgClosure *caf_list;
 
+extern StgTSO *resurrected_threads;
+
 extern bdescr *small_alloc_list;
 extern bdescr *large_alloc_list;
 extern bdescr *pinned_object_block;
@@ -50,6 +55,7 @@ extern void     allocNurseries ( void );
 extern void     resetNurseries ( void );
 extern bdescr * allocNursery   ( bdescr *last_bd, nat blocks );
 extern void     resizeNursery  ( nat blocks );
+extern void     tidyAllocateLists ( void );
 
 /* Stats 'n' stuff */
 extern lnat calcAllocated  ( void );
@@ -79,16 +85,6 @@ extern void checkSanity(void);
 extern nat  countBlocks(bdescr *);
 #endif
 
-/* 
- * These three are used by the garbage collector when we have
- * dynamically-linked object modules.  (see ClosureMacros.h,
- * IS_CODE_PTR etc.). 
- * Defined in Linker.c.
- */
-int is_dynamically_loaded_code_or_rodata_ptr ( void* p );
-int is_dynamically_loaded_rwdata_ptr         ( void* p );
-int is_not_dynamically_loaded_ptr            ( void* p );
-
 /* Functions from GC.c 
  */
 extern void         threadPaused ( StgTSO * );
