@@ -11,12 +11,9 @@ module Layout ( LayoutItem(..), Root(..), Relative
 
 
 import Prelude hiding (Either(..))	-- Added to hide Left and Right
-import CleanStdList
-import CleanStdMisc
-import Commondef
+import CommonDef
 import Id
-import Windowhandle(LayoutInfo(..), Origin)
-
+import WindowHandle(LayoutInfo(..), Origin)
 
 layoutError :: String -> String -> x
 layoutError rule message
@@ -67,13 +64,13 @@ removeRoot :: Id -> [Root] -> (Bool,Root,[Root])
 removeRoot id (item:items)
 	| identifyRoot id item
 		= (True,item,items)
-	| contains (identifyRelative id) (rootTree item)
+	| any (identifyRelative id) (rootTree item)
 		= (True,item,items)
 	| otherwise
 		= let (found,root,items1) = removeRoot id items
 		  in  (found,root,item:items1)
 removeRoot _ items
-	= (False,undef,items)
+	= (False,undefined,items)
 
 
 {-	getLayoutItemPosSize id retrieves the position and size of:
@@ -89,7 +86,7 @@ getLayoutItemPosSize id item
 	| otherwise
 		= (True,relativePos relative,liItemSize $ relativeItem $ relative)
 	where
-		(found,relative) = cselect (identifyRelative id) undef (rootTree item)
+		(found,relative) = cselect (identifyRelative id) undefined (rootTree item)
 
 
 {-	shiftRelative v shifts the position of the Relative argument by v.
@@ -133,7 +130,7 @@ getRootBoundingBox (Root {rootPos=rootPos,rootItem=rootItem,rootTree=rootTree})
 getLayoutItem :: Id -> [Root] -> (LayoutInfo,Vector2,Size,[Root])
 getLayoutItem id items@(root:roots)
 	| identifyRoot id root
-		= if   isEmpty depends
+		= if   null depends
 		  then (layoutInfo,corner,size,roots)
 		  else (layoutInfo,corner,size,items)
 	| inTree
@@ -177,9 +174,9 @@ layoutItems hMargins@(lMargin,rMargin) vMargins@(tMargin,bMargin) itemSpaces req
 	where
 		reqSize1     = if reqSize/=zero then Size {w=w reqSize-lMargin-rMargin,h=h reqSize-tMargin-bMargin} else reqSize
 		layoutItems1 = sortLayoutItems layoutItems
-		(_,roots)    = stateMap2 (calcRootPosition itemSpaces orientations) layoutItems1 (0,[])
+		(_,roots)    = foldr (calcRootPosition itemSpaces orientations) (0,[]) layoutItems1
 		size         = calcAreaSize orientations roots reqSize1 minSize
-		roots1       = stateMap2 (calcFramePosition hMargins vMargins orientations size) roots []
+		roots1       = foldr (calcFramePosition hMargins vMargins orientations size) [] roots
 		finalSize    = Size {w=lMargin+w size+rMargin,h=tMargin+h size+bMargin}
 
 
@@ -197,16 +194,14 @@ sortLayoutItems layoutItems
 		(lineItems,relItems) = divide (\item->isLine (liItemPos item)) layoutItems
 		
 		sortLayoutItems' :: [LayoutItem] -> [LayoutItem] -> [LayoutItem]
-		sortLayoutItems' done todo
-			| isEmpty todo
-				= reverse done
+		sortLayoutItems' done [] = done
+		sortLayoutItems' done (item1:todo)
 			| not isrelative
-				= sortLayoutItems' (item1:done) todo1
+				= sortLayoutItems' (item1:done) todo
 			| otherwise
-				= let (done1,chain,todo2) = getItemPosChain id2 done [item1] todo1
-				  in  sortLayoutItems' (insertItemPosChain chain done1) todo2
+				= let (done1,chain,todo1) = getItemPosChain id2 done [item1] todo
+				  in  (sortLayoutItems' (insertItemPosChain chain done1) todo1)
 			where
-				(item1,todo1)    = hdtl todo
 				pos1             = liItemPos item1
 				(isrelative,id2) = isRelative pos1
 				
@@ -223,9 +218,9 @@ sortLayoutItems layoutItems
 					| otherwise
 						= getItemPosChain id2 done (next:chain) todo1
 					where
-						in_chain             = contains (identifyLayoutItem nextId) chain
-						in_done              = contains (identifyLayoutItem nextId) done
-						(in_todo,next,todo1) = remove (identifyLayoutItem nextId) undef todo
+						in_chain             = any (identifyLayoutItem nextId) chain
+						in_done              = any (identifyLayoutItem nextId) done
+						(in_todo,next,todo1) = remove (identifyLayoutItem nextId) undefined todo
 						nextPos              = liItemPos next
 						(isrelative,id2)     = isRelative nextPos
 				
@@ -263,7 +258,7 @@ sortLayoutItems layoutItems
 calcRootPosition :: (Int,Int) -> [(ViewDomain,Origin)] -> LayoutItem -> (Int,[Root]) -> (Int,[Root])
 calcRootPosition itemSpaces orientations item1 sDone@(sizeY,done)
 	| isfix
-		= let	(_,origin) = hd orientations
+		= let	(_,origin) = head orientations
 			itemoffset = itemPosOffset fixpos orientations
 			pos        = itemoffset-toVector origin
 			item1'     = Root {rootItem=item1,rootPos=pos,rootTree=[]}
@@ -351,9 +346,9 @@ calcAreaSize orientations roots reqSize minimumSize
 	| reqSize/=zero
 		= stretchSize minimumSize reqSize
 	| otherwise
-		= stateMap2 (fitRootInArea origin orientations) roots minimumSize
+		= foldr (fitRootInArea origin orientations) minimumSize roots
 	where
-		origin = snd (hd orientations)
+		origin = snd (head orientations)
 		
 		stretchSize :: Size -> Size -> Size
 		stretchSize size1 size2 = Size {w=max (w size1) (w size2), h=max (h size1) (h size2)}
