@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: HsUtil.h,v 1.1.6.1 2001/12/13 12:10:00 simonmar Exp $
+ * $Id: HsUtil.h,v 1.2.16.1 2003/09/16 12:04:02 simonmar Exp $
  *
  * Definitions for package `util' which are visible in Haskell land.
  *
@@ -8,11 +8,6 @@
 #ifndef HSUTIL_H
 #define HSUTIL_H
 
-/* md5.c */
-extern void __hslibs_MD5Init(StgByteArray context);
-extern void __hslibs_MD5Update(StgByteArray context, void const *buf, int len);
-extern void __hslibs_MD5Final(StgByteArray digest, StgByteArray context);
-
 /* selectFrom.c */
 extern StgInt sizeof_fd_set__();
 extern void fd_zero__(StgByteArray fds);
